(******************************************************************************

______________________________________________________________________________

ColBuild v0.10                                               (c) 2009-12 Pepak
http://www.pepak.net/e-books/prs-505-sbirky/              http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2009-12 Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uDevices;

interface

uses
  SysUtils, Classes, Windows,
  uEbooks, uMedia;

type
  EDeviceError = class(Exception);

type
  TDeviceClass = class of TDevice;
  TDevice = class
    private
      fRootDir: string;
      fExpandedRootDir: string;
      fExpandedRootDirLength: integer;
      fConnected: boolean;
      fMediaType: TMediaType;
      fDevicePathDelimiter: string;
      procedure SetConnected(const Value: boolean);
      procedure SetMediaType(const Value: TMediaType);
    protected
      function SupportedMediaTypes: TMediaTypes; virtual;
      function PathDelimiterToDevicePathDelimiter(const s: string): string;
      function DevicePathDelimiterToPathDelimiter(const s: string): string;
      function FindFilePathInReader(const FileName: string): string;
      property RootDir: string read fRootDir;
      property DevicePathDelimiter: string read fDevicePathDelimiter write fDevicePathDelimiter;
    public
      class procedure RegisterDeviceClass(const Shortcut: string; AClass: TDeviceClass);
      class procedure ListDevices(L: TStrings);
      class function CreateDevice(const Shortcut, RootDir: string; out Device: TDevice): boolean;
      class function Description: string; virtual; abstract;
      constructor Create(const ARootDir: string); virtual;
      destructor Destroy; override;
      procedure Connect; virtual;
      procedure Disconnect; virtual;
      property Connected: boolean read fConnected write SetConnected;
      property MediaType: TMediaType read fMediaType write SetMediaType;
      procedure Commit; virtual;
      procedure Rollback; virtual;
      function IsSupportedFile(const AFileName: string): boolean; virtual; abstract;
      function IsRegisteredBook(Ebook: TEbookFile): boolean; virtual; abstract;
      function DeleteAllCollections: boolean; virtual; abstract;
      function RegisterBook(Ebook: TEbookFile): boolean; virtual; abstract;
      function RegisterBooks(Ebooks: TEbookFileList; Failed: TList): boolean;
      function CreateCollection(const Name, NameSort: string; Ebooks: TEbookFileList): boolean; virtual; abstract;
      function FindFilePathRelativeToRoot(Ebook: TEbookFile): string;
    end;

implementation

{ TDevice }

var
  DeviceClasses: TStringList;

class procedure TDevice.RegisterDeviceClass(const Shortcut: string; AClass: TDeviceClass);
begin
  if AClass <> nil then
    DeviceClasses.AddObject(Shortcut, TObject(AClass));
end;

class procedure TDevice.ListDevices(L: TStrings);
{
var
  i: integer;
}
begin
{
  L.Clear;
  L.Capacity := DeviceClasses.Count;
  for i := 0 to Pred(DeviceClasses.Count) do
    L.Add(DeviceClasses[i]);
}
  L.Assign(DeviceClasses);
end;

class function TDevice.CreateDevice(const Shortcut, RootDir: string; out Device: TDevice): boolean;
var
  ix: integer;
begin
  if DeviceClasses.Find(Shortcut, ix) then
    begin
    Device := TDeviceClass(DeviceClasses.Objects[ix]).Create(RootDir);
    Result := True;
    end
  else
    begin
    Device := nil;
    Result := False;
    end;
end;

constructor TDevice.Create(const ARootDir: string);
begin
  inherited Create;
  fDevicePathDelimiter := '/';
  fRootDir := ARootDir;
  fExpandedRootDir := PathDelimiterToDevicePathDelimiter(IncludeTrailingPathDelimiter(ExpandFileName(ARootDir)));
  fExpandedRootDirLength := Length(fExpandedRootDir);
end;

destructor TDevice.Destroy;
begin
  Disconnect;
  inherited;
end;

procedure TDevice.Connect;
begin
  fConnected := True;
end;

procedure TDevice.Disconnect;
begin
  fConnected := False;
end;

function TDevice.FindFilePathInReader(const FileName: string): string;
begin
  Result := DevicePathDelimiterToPathDelimiter(fExpandedRootDir + FileName);
end;

function TDevice.FindFilePathRelativeToRoot(Ebook: TEbookFile): string;
begin
  Result := PathDelimiterToDevicePathDelimiter(ExpandFileName(Ebook.FileName));
  if AnsiCompareText(fExpandedRootDir, Copy(Result, 1, fExpandedRootDirLength)) <> 0 then
    Raise EInOutError.CreateFmt('Invalid filename "%s" (outside of rootdir)', [Ebook.FileName]);
  Delete(Result, 1, fExpandedRootDirLength);
end;

procedure TDevice.Commit;
begin
  // To be overridden
end;

procedure TDevice.Rollback;
begin
  // To be overridden
end;

procedure TDevice.SetConnected(const Value: boolean);
begin
  if fConnected <> Value then
    if Value then
      Connect
    else
      Disconnect;
end;

procedure TDevice.SetMediaType(const Value: TMediaType);
var
  s: string;
  mt: TMediaType;
begin
  if Value in SupportedMediaTypes then
    fMediatype := Value
  else
    begin
    s := '';
    for mt := Low(TMediaType) to High(TMediaType) do
      if mt in SupportedMediaTypes then
        if s = '' then
          s := MediaTypeInfo[mt].Description
        else
          s := s + ', ' + MediaTypeInfo[mt].Description;
    Raise EDeviceError.CreateFmt('Unsupported media type %s. Device %s supports %s', [MediaTypeInfo[Value].Description, Description, s]);
    end;
end;

function TDevice.PathDelimiterToDevicePathDelimiter(const s: string): string;
begin
  Result := StringReplace(s, '\', DevicePathDelimiter, [rfReplaceAll]);
  Result := StringReplace(s, PathDelim, DevicePathDelimiter, [rfReplaceAll]);
end;

function TDevice.DevicePathDelimiterToPathDelimiter(const s: string): string;
begin
  Result := StringReplace(s, DevicePathDelimiter, PathDelim, [rfReplaceAll]);
end;

function TDevice.SupportedMediaTypes: TMediaTypes;
begin
  Result := [Low(TMediaType)..High(TMediaType)];
end;

function TDevice.RegisterBooks(Ebooks: TEbookFileList; Failed: TList): boolean;
var
  i: Integer;
begin
  Result := True;
  for i := 0 to Pred(Ebooks.Count) do
    if not RegisterBook(Ebooks[i]) then
      begin
      Failed.Add(Ebooks[i]);
      Result := False;
      end;
end;

initialization
  DeviceClasses := TStringList.Create;
  DeviceClasses.Sorted := True;
  DeviceClasses.Duplicates := dupError;

finalization
  FreeAndNil(DeviceClasses);

end.
