(******************************************************************************

______________________________________________________________________________

ColBuild v0.10                                               (c) 2009-12 Pepak
http://www.pepak.net/e-books/prs-505-sbirky/              http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2009-12 Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uPRS_T1;

interface

uses
  SysUtils, Classes, Windows, Graphics, Jpeg,
  SQLiteTable3,
  uDevices, uMedia, uEbooks, uSystem, uGuid;

type
  TSonyPRST1 = class(TDevice)
    private
    protected
      Database: TSQLiteDatabase;
      function Str2Sql(const s: string; NullOnEmpty: boolean = True): string;
      function DateTime2Sql(const d: TDateTime; NullOnEmpty: boolean = True): string;
      function FindBookID(Ebook: TEbookFile; out ID: int64): boolean;
      function GetMimeType(Ebook: TEbookFile): string;
      function UpdateBook(Ebook: TEbookFile; ID: int64): boolean;
      function CreateBook(Ebook: TEbookFile; out ID: int64): boolean;
      function UpdateBookThumbnail(Ebook: TEbookFile; ID: int64): boolean;
      function GetCleanDatabaseResourceName: string; virtual;
      procedure BuildCleanDatabase; virtual;
      function SupportedMediaTypes: TMediaTypes; override;
      function SourceID: string;
    public
      class function Description: string; override;
      constructor Create(const ARootDir: string); override;
      destructor Destroy; override;
      procedure Connect; override;
      procedure Disconnect; override;
      procedure Commit; override;
      procedure Rollback; override;
      function IsSupportedFile(const AFileName: string): boolean; override;
      function IsRegisteredBook(Ebook: TEbookFile): boolean; override;
      function DeleteAllCollections: boolean; override;
      function RegisterBook(Ebook: TEbookFile): boolean; override;
      function CreateCollection(const Name, NameSort: string; Ebooks: TEbookFileList): boolean; override;
    end;

implementation

{$R 'Resources\PRS_T1_BooksDB.res'}

const
  FILENAME_BOOKSDB = 'Sony_Reader\database\books.db';
  FILENAME_THUMBNAIL_BY_ID = 'Sony_Reader\database\cache\books\%d\thumbnail\main_thumbnail.jpg';

type
  TMimeType = record
    Extension, MimeType: string;
    end;
const
  MimeTypes: array[0..2] of TMimeType = (
    (Extension: '.epub'; MimeType: 'application/epub+zip'),
    (Extension: '.txt'; MimeType: 'text/plain'),
    (Extension: '.pdf'; MimeType: 'application/pdf')
  );

{ TSonyPRST1 }

class function TSonyPRST1.Description: string;
begin
  Result := 'Sony Reader PRS-T1';
end;

constructor TSonyPRST1.Create(const ARootDir: string);
begin
  inherited;
end;

destructor TSonyPRST1.Destroy;
begin
  inherited;
end;

procedure TSonyPRST1.Connect;
var
  FN: string;
begin
  if not Connected then
    begin
    FN := FindFilePathInReader(FILENAME_BOOKSDB);
    if not FileExists(FN) then
      BuildCleanDatabase;
    Database := TSQLiteDatabase.Create(FN);
    Database.BeginTransaction;
    end;
  inherited;
end;

procedure TSonyPRST1.Disconnect;
begin
  if Connected then
    begin
    Rollback;
    FreeAndNil(Database);
    end;
  inherited;
end;

procedure TSonyPRST1.Commit;
begin
  Database.Commit;
  Database.BeginTransaction;
  inherited;
end;

procedure TSonyPRST1.Rollback;
begin
  Database.Rollback;
  Database.BeginTransaction;
  inherited;
end;

function TSonyPRST1.Str2Sql(const s: string; NullOnEmpty: boolean): string;
begin
  if NullOnEmpty and (s = '') then
    Result := 'NULL'
  else
    Result := AnsiQuotedStr(s, '''');
end;

function TSonyPRST1.SupportedMediaTypes: TMediaTypes;
begin
  Result := [mtInternalMemory, mtSDcard];
end;

function TSonyPRST1.DateTime2Sql(const d: TDateTime; NullOnEmpty: boolean): string;
const
  UNIXTIME_START = 25569;
var
  SysTime: TSystemTime;
  d1, d2: TDateTime;
  dd: int64;
begin
  if NullOnEmpty and (d <= 0) then
    Result := 'NULL'
  else
    begin
    GetSystemTime(SysTime);
    d1 := Date;
    d2 := EncodeDate(SysTime.wYear, SysTime.wMonth, SysTime.wDay);
    dd := Trunc((d - d1 + d2 - UNIXTIME_START)*24*60*60*1000 + 0.5);
    Result := IntToStr(dd);
    end;
end;

function TSonyPRST1.DeleteAllCollections: boolean;
begin
  Connect;
  Database.ExecSQL('DELETE FROM collection');
  Result := True;
end;

function TSonyPRST1.IsRegisteredBook(Ebook: TEbookFile): boolean;
var
  Dummy: int64;
begin
  Result := FindBookID(Ebook, Dummy);
end;

function TSonyPRST1.IsSupportedFile(const AFileName: string): boolean;
var
  Ext: string;
  i: integer;
begin
  Result := False;
  Ext := ExtractFileExt(AFileName);
  for i := 0 to Pred(Length(MimeTypes)) do
    if AnsiCompareText(Ext, MimeTypes[i].Extension) = 0 then
      begin
      Result := True;
      Break;
      end;
end;

function TSonyPRST1.SourceID: string;
begin
  case MediaType of
    mtInternalMemory:
      Result := '0';
    mtSDcard:
      Result := '1';
    else
      Result := '';
  end;
end;

function TSonyPRST1.GetCleanDatabaseResourceName: string;
begin
  Result := 'PRS_T1_BooksDB';
end;

procedure TSonyPRST1.BuildCleanDatabase;
var
  FN: string;
  DBStream, ResStream: TStream;
begin
  FN := FindFilePathInReader(FILENAME_BOOKSDB);
  ForceDirectories(ExtractFilePath(FN));
  ResStream := TResourceStream.Create(hInstance, GetCleanDatabaseResourceName, 'BINARY');
  try
    DBStream := TFileStream.Create(FN, fmCreate);
    try
      DBStream.CopyFrom(ResStream, 0);
    finally
      FreeAndNil(DBStream);
      end;
  finally
    FreeAndNil(ResStream);
    end;
end;

function TSonyPRST1.GetMimeType(Ebook: TEbookFile): string;
var
  Ext: string;
  i: integer;
begin
  Result := '';
  Ext := ExtractFileExt(Ebook.FileName);
  for i := 0 to Pred(Length(MimeTypes)) do
    if AnsiCompareText(Ext, MimeTypes[i].Extension) = 0 then
      begin
      Result := MimeTypes[i].MimeType;
      Break;
      end;
end;

function TSonyPRST1.FindBookID(Ebook: TEbookFile; out ID: int64): boolean;
var
  FN: string;
begin
  FN := FindFilePathRelativeToRoot(Ebook);
  Connect;
  ID := Database.GetTableValue('SELECT _id FROM books WHERE file_path=' + Str2Sql(FN));
  Result := ID <> 0;
end;

function TSonyPRST1.CreateBook(Ebook: TEbookFile; out ID: int64): boolean;
var
  Sql: string;
begin
  // datumy: (unix timestamp v GMT) * 1000 + miliseconds
  Sql := Format('INSERT INTO books ('
    + ' title, author, kana_title, kana_author, title_key, author_key,'
    + ' source_id, added_date, modified_date, file_path, file_name, file_size,'
    + ' mime_type, corrupted, prevent_delete'
    + ') VALUES ('
    + ' %s, %s, %s, %s, %s, %s,'
    + ' %s, %s, %s, %s, %s, %d,'
    + ' %s, 0, 0'
    + ')',
    [
     Str2Sql(Ebook.Title), Str2Sql(Ebook.Author), Str2Sql(Ebook.TitleSort), Str2Sql(Ebook.AuthorSort), Str2Sql(Ebook.TitleSort), Str2Sql(Ebook.AuthorSort),
     Str2Sql(SourceID), DateTime2Sql(Now), DateTime2Sql(Ebook.ModificationDate), Str2Sql(FindFilePathRelativeToRoot(Ebook)), Str2Sql(ExtractFileName(Ebook.FileName)), Ebook.Size,
     Str2Sql(GetMimeType(Ebook))
    ]);
  Connect;
  Database.ExecSql(Sql);
  ID := Database.GetLastInsertRowID;
  Result := True;
end;

function TSonyPRST1.CreateCollection(const Name, NameSort: string; Ebooks: TEbookFileList): boolean;
var
  Sql: string;
  IDCol, IDBook: int64;
  i: integer;
begin
  Sql := Format('INSERT INTO collection ('
    + ' title, kana_title, source_id, uuid'
    + ') VALUES ('
    + ' %s, %s, %s, %s'
    + ')',
    [
     Str2Sql(Name), Str2Sql(NameSort), Str2Sql(SourceID), Str2Sql(GenerateUuid)
    ]);
  Connect;
  Database.ExecSql(Sql);
  IDCol := Database.GetLastInsertRowID;
  for i := 0 to Pred(Ebooks.Count) do
    if FindBookID(Ebooks[i], IDBook) or CreateBook(Ebooks[i], IDBook) then
      begin
      Sql := Format('INSERT INTO collections ('
        + ' collection_id, content_id, added_order'
        + ') VALUES ('
        + ' %d, %d, %d'
        + ')',
        [
          IDCol, IDBook, Succ(i)
        ]);
      Database.ExecSql(Sql);
      end;
  Result := True;
end;

function TSonyPRST1.UpdateBook(Ebook: TEbookFile; ID: int64): boolean;
var
  Sql: string;
begin
  Sql := Format('UPDATE books'
    + ' SET title=COALESCE(%s, title)'
    + '     ,author=COALESCE(%s, author)'
    + '     ,kana_title=COALESCE(%s, kana_title)'
    + '     ,kana_author=COALESCE(%s, kana_author)'
    + '     ,title_key=COALESCE(%s, title_key)'
    + '     ,author_key=COALESCE(%s, author_key)'
    //+ '     ,source_id'
    //+ '     ,added_date'
    //+ '     ,modified_date=COALESCE(%s, modified_date)'
    //+ '     ,file_path'
    //+ '     ,file_name'
    //+ '     ,file_size=%d'
    //+ '     ,mime_type'
    //+ '     ,corrupted'
    //+ '     ,prevent_delete'
    + ' WHERE _id=%d',
    [
     Str2Sql(Ebook.Title), Str2Sql(Ebook.Author), Str2Sql(Ebook.TitleSort), Str2Sql(Ebook.AuthorSort), Str2Sql(Ebook.TitleSort), Str2Sql(Ebook.AuthorSort),
     {DateTime2Sql(Now), DateTime2Sql(Ebook.ModificationDate), Str2Sql(FindFilePath(Ebook)), Str2Sql(ExtractFileName(Ebook.FileName)), Ebook.Size,}
     {GetMimeType(Ebook)}
     ID
    ]);
  Connect;
  Database.ExecSql(Sql);
  Result := True;
end;

function TSonyPRST1.UpdateBookThumbnail(Ebook: TEbookFile; ID: int64): boolean;
var
  Sql, Thumbnail: string;
  Bitmap: TBitmap;
  Thumb: TJpegImage;
begin
  Result := False;
  if Ebook.Cover <> nil then
    begin
    Thumbnail := Database.GetTableString(Format('SELECT thumbnail FROM books WHERE _id=%d', [ID]));
    if Thumbnail <> '' then
      Thumbnail := DevicePathDelimiterToPathDelimiter(Thumbnail)
    else
      Thumbnail := Format(FILENAME_THUMBNAIL_BY_ID, [ID]);
    Bitmap := TBitmap.Create;
    try
      Bitmap.SetSize(112, 144);
      Bitmap.Canvas.StretchDraw(Rect(0, 0, Pred(Bitmap.Width), Pred(Bitmap.Height)), Ebook.Cover);
      Thumb := TJpegImage.Create;
      try
        Thumb.Assign(Bitmap);
        ForceDirectories(ExtractFilePath(FindFilePathInReader(Thumbnail)));
        Thumb.SaveToFile(FindFilePathInReader(Thumbnail));
        Sql := Format('UPDATE books SET thumbnail=COALESCE(%s, thumbnail) WHERE _id=%d', [Str2Sql(PathDelimiterToDevicePathDelimiter(Thumbnail)), ID]);
        Connect;
        Database.ExecSql(Sql);
        Result := True;
      finally
        FreeAndNil(Thumb);
        end;
    finally
      FreeAndNil(Bitmap);
      end;
    end;
end;

function TSonyPRST1.RegisterBook(Ebook: TEbookFile): boolean;
var
  ID: int64;
begin
  if FindBookID(Ebook, ID) then
    Result := UpdateBook(Ebook, ID)
  else
    Result := CreateBook(Ebook, ID);
  if Result then
    UpdateBookThumbnail(Ebook, ID);
end;

initialization
  TDevice.RegisterDeviceClass('prs-t1', TSonyPRST1);

end.

